# -*- coding: utf-8 -*-
from pyrfc3339 import parse
import MySQLdb
import MySQLdb.cursors as cursors

# Backup idea -- oracle's connector:
#import mysql.connector

#TODO: Streaming result set for very large returns
#      https://techualization.blogspot.com/2011/12/retrieving-million-of-rows-from-mysql.html

from ..log import log

DEFAULT_HOST = 'localhost'
DEFAULT_USER = 'medgen'
DEFAULT_PASS = 'medgen'
DEFAULT_DATASET = 'medgen'

SQLDATE_FMT = '%Y-%m-%d %H:%M:%S'
def EscapeString(value):
    value = value.encode("utf-8")
    #value = MySQLdb.escape_string(value)
    return '"{}"'.format(value)

def SQLdatetime(pydatetime_or_string):
    if hasattr(pydatetime_or_string, 'strftime'):
        dtobj = pydatetime_or_string
    else:
        # assume pyrfc3339 string
        dtobj = parse(pydatetime_or_string)
    return dtobj.strftime(SQLDATE_FMT)

class SQLData(object):
    """
    MySQL base class for config, select, insert, update, and delete of medgen linked databases.

    See https://dev.mysql.com/doc/connector-python/en/connector-python-example-connecting.html
    """
    def __init__(self, *args, **kwargs):
        self._cfg_section = kwargs.get('config_section', 'DEFAULT')

        from ..config import config
        self._db_host = kwargs.get('db_host', None) or config.get(self._cfg_section, 'db_host')
        self._db_user = kwargs.get('db_user', None) or config.get(self._cfg_section, 'db_user')
        self._db_pass = kwargs.get('db_pass', None) or config.get(self._cfg_section, 'db_pass')
        self._db_name = kwargs.get('dataset', None) or config.get(self._cfg_section, 'dataset')

        # will be connected upon first query, or can be set up "manually" by doing self.connect()
        self.conn = None

        #DEMOLITION / cruft from PySQLPool
        #self.commitOnEnd = kwargs.get('commitOnEnd', True) or config.get(self._cfg_section, 'commitOnEnd')

    def connect(self):
        self.conn = MySQLdb.connect(passwd=self._db_pass,
                                    user=self._db_user,
                                    db=self._db_name,
                                    host=self._db_host,
                                    cursorclass=cursors.DictCursor,
                                    charset='utf8',
                                    use_unicode=True,
                                   )
        return self.conn

    def cursor(self, execute_sql=None):
        if not self.conn:
            self.connect()
        cursor = self.conn.cursor(cursors.DictCursor)

        if execute_sql is not None:
            cursor.execute(execute_sql)

        return cursor

    def fetchall(self, select_sql, *args):
        """ For submitted select_sql with interpolation strings meant to match
        with supplied *args, build and execute the statement and fetch all results.

        Results will be returned as a list of dictionaries.

        Example:
            DB.fetchall('select HGVS from clinvar where PMID="%s"', ('21129721',))

        :param select_sql: (str)
        :returns: results as list of dictionaries
        :rtype: list
        """
        # this line opens a cursor, executes, gets the data, and closes the cursor.
        if args:
            results = self.cursor(select_sql % args).fetchall()
        else:
            results = self.cursor(select_sql).fetchall()
        return results

    def fetchrow(self, select_sql, *args):
        """
        If the query was successful:
            if 1 or more rows was returned, returns the first one
            else returns None
        Else:
            raises Exception
        """
        res = self.fetchall(select_sql, *args)
        return res[0] if len(res) > 0 else None

    #def fetchID(self, select_sql, id_colname='ID', *args, **kwargs):
    def fetchID(self, select_sql, *args, **kwargs):
        id_colname = kwargs.get('id_colname', 'ID')

        results = self.fetchrow(select_sql, *args)
        if results is not None:
            if id_colname in results:
                return results[id_colname]
            else:
                raise RuntimeError("No ID column found.  SQL query: %s" % select_sql)
        return None  # no results found

    # DEMOLITION / Probably not needed
    #def list_concepts(self, select_sql):
    #    """
    #    Fetch list of concepts
    #    :param select_sql: query
    #    :return: list cui
    #    """
    #    return self.fetchlist(select_sql, 'CUI')

    # DEMOLITION / Probably not needed
    #def list_genes(self, select_sql):
    #    """
    #    Fetch list of genes
    #    :param select_sql: query
    #    :return: list HGNC
    #    """
    #   return self.fetchlist(select_sql, 'gene_name')

    def insert(self, tablename, field_value_dict):
        """ Takes a tablename (should be found in selected DB) and a dictionary mapping
        column and value pairs, and inserts the dictionary as a row in target table.

        WARNING: this function trusts field names (not protected against sql injection attacks).

        :param: tablename: name of table to receive new row
        :param: field_value_dict: map of field=value
        :return: row_id (integer) (returns 0 if insert failed)
        """
        fields = []
        values = []

        for key, val in field_value_dict.items():
            fields.append(key)
            values.append(val)

        sql = 'insert into {} ({}) values ({});'.format(tablename, ','.join(fields), ','.join(['%s' for v in values]))
        cursor = self.execute(sql, *values)
        cursor.close()
        return self.conn.insert_id()

    def update(self, tablename, id_col_name, row_id, field_value_dict):
        """
        :param: tablename: name of table to update
        :param: row_id (int): row id of record to update
        :param: field_value_dict: map of field=value
        :return: row_id (integer) (returns 0 if insert failed)
        """
        fields = []
        values = []

        clauses = []

        for key, val in field_value_dict.items():
            clause = '%s=' % key
            # surround strings and datetimes with quotation marks
            if val == None:
                clause += 'NULL'
            elif hasattr(val, 'strftime'):
                clause += '"%s"' % val.strftime(SQLDATE_FMT)
            elif hasattr(val, 'lower'):
                clause += EscapeString(val)  #surrounds strings with quotes and unicodes them.
            else:
                clause += str(val)
            clauses.append(clause)

        sql = 'update '+tablename+' set %s where %s=%i;' % (', '.join(clauses), id_col_name, row_id)
        queryobj = self.execute(sql)
        # retrieve and return the row id of the insert. returns 0 if insert failed.
        return queryobj.lastInsertID

    def delete(self, tablename, field_value_dict):
        """
        :param: tablename: name of table from which rows will be deleted.
        :param: field_value_dict: map of field=value indicating which rows to delete.
        :return: row_id (integer) (returns 0 if delete failed)
        """
        if len(field_value_dict) == 0:
            raise RuntimeError("Do not support delete without a WHERE clause")

        where_sql = ''
        for key, val in field_value_dict.items():
            if val == None:
                val = 'NULL'
                where_sql += 'AND {} is NULL '.format(key)
            # surround strings and datetimes with quotation marks
            elif hasattr(val, 'strftime'):
                val = '"%s"' % val.strftime(SQLDATE_FMT)
                where_sql += 'AND {}={} '.format(key, val)
            elif hasattr(val, 'lower'):
                val = EscapeString(val)  # surrounds strings with quotes and unicodes them.
                where_sql += 'AND {}={} '.format(key, val)
            else:
                val = str(val)
                where_sql += 'AND {}={} '.format(key, val)

        where_sql = where_sql[len('AND '):]

        sql = 'delete from {} where {}'.format(tablename, where_sql)

        cursor = self.execute(sql)
        #TODO: does this do anything? work at all?? eek.
        result = cursor.fetchone()
        cursor.close()
        return result

    def drop_table(self, tablename):
        cursor = self.execute('drop table if exists ' + tablename)
        result = cursor.fetchone()
        cursor.close()
        return result

    def truncate(self, tablename):
        cursor = self.execute('truncate ' + tablename)
        result = cursor.fetchone()
        cursor.close()
        return result

    def execute(self, sql, *args):
        """
        Executes arbitrary sql string in current database connection.
        Use this method's args for positional string interpolation into sql.

        :param sql: (str)
        :return: MySQLdb cursor object
        """
        #try:
        log.debug('SQL.execute ' + sql, *args)
        #except TypeError as error:
        #    log.info('Arguments do not match number of string interpolations in SQL statement.')
        #    log.info('  SQL was: %s' % sql)
        #    log.info('  Args: %r' % args)

        try:
            cursor = self.cursor(sql % args)
        except Exception as err:
            log.info('Medgen SQL ERROR: %r' % err)
            full_sql = sql % args
            log.info('Tripped on a piece of SQL: ' + full_sql)
        return cursor 

    def ping(self):
        """
        Same effect as calling 'mysql> call mem'
        :returns::self.schema_info(()
        """
        try:
            return self.schema_info()
        except MySQLdb.Error as err:
            log.error('DB connection is dead %d: %s' % (err.args[0], err.args[1]))
            return False

    def schema_info(self):
        header = ['schema', 'engine', 'table', 'rows', 'million', 'data length', 'MB', 'index']
        return {'header': header, 'tables': self.fetchall('call mem')}

    def last_loaded(self, dbname=None):
        if dbname is None:
            dbname = self._db_name
        return self.fetchID('select event_time as ID from '+dbname+'.log where entity_name = "load_database.sh" and message = "done" order by idx desc limit 1')

    def unique_PMID(self, sql):
        """
        For given sql select query, return a list of unique PMID strings.
        """
        pubmeds = set()
        for row in self.fetchall(sql):
            pubmeds.add(str(row['PMID']))
        return pubmeds

    def unique_HGVS(self, sql):
        """
        For given sql select query, return a list of unique HGVS strings.
        """
        hgvs_texts = set()
        for row in self.fetchall(sql):
            hgvs_texts.add(str(row['HGVS']))
        return hgvs_texts

    def trunc_str(self, inp, maxlen):
        """
        Useful utility method for storing text in a database
        :param inp: a string
        :param maxlen: the max length of that string
        :return: '%s' % s or '%s...' % s[:m-3]
        """
        if maxlen < 3:
            raise RuntimeError('maxlen must be at least 3')
        if len(inp) > maxlen:
            inp = '%s...' % inp[:maxlen - 3]
        return inp

    def get_last_mirror_time(self, entity_name):
        """
        Get the last time some entity was mirrored

        :param entity_name: table name for db, for example, bic_brca1
        :return: datetime if found
        """
        sql_query = 'SELECT event_time FROM log WHERE entity_name = "%s" AND message like "rows loaded %s" ORDER BY event_time DESC limit 1'
        result = self.fetchrow(sql_query, entity_name, '%')
        if result:
            return result['event_time']
        sql_query = sql_query % (entity_name, '%')
        raise RuntimeError('Query "%s" returned no results. Have you loaded the %s table?' % (sql_query, entity_name))

    def create_index(self, table, colspec):
        """
        Create index on a specified table using the colums defined.
        Index start/stop times are logged to the "log" table.
        :param table: name of the table, example, "train"
        :param colspec: name of column, for example, "RQ"
        :return:
        """
        self.execute("call create_index(%s, %s) ", (table, colspec))

    def fetchlist(self, select_sql, column='gene_name'):
        """
        Fetch as list
        :param select_sql: query
        :param column: name of column you want to make a list out of
        :return: list
        """
        rows = self.fetchall(select_sql)
        return [] if rows is None else [str(r[column]) for r in rows]



__doc__= """ Configuration setup """

import os
from configparser import ConfigParser
from .log import log

# TODO: documentation?

PKGNAME = 'medgen'

# env variable MEDGEN_CONFIG_DIR should be an absolute path to a directory.
# if this env doesn't exist, default to looking in package's config files.

default_config_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config')

CFGDIR = os.getenv('MEDGEN_CONFIG_DIR', default_config_dir)
ENV = os.getenv('MEDGEN_ENV', 'default')


log.info("MedGen Configuration Settings: ")
log.info("CFGDIR: %s" % CFGDIR)
log.info("ENV: %s" % ENV)


def _get_config(dirname=CFGDIR, env=ENV):
    config = ConfigParser()
    configs = [os.path.join(dirname, x) for x in os.listdir(dirname) if x.find(env) > -1]
    config.read(configs)
    return config

config = _get_config()




import requests, json, urllib

from metapub.text_mining import is_pmcid, is_ncbi_bookID 
from metapub.pubmedcentral import get_pmid_for_otherid

from ..db.clinvar import ClinVarDB
from ..log import log, IS_DEBUG_ENABLED

##########################################################################################
#
#   Functions
#
##########################################################################################


def _ncbi_variant_report_service(hgvs_text):
    """
    Return results from API query to the NCBI Variant Reporter Service
    See documentation at:
    http://www.ncbi.nlm.nih.gov/variation/tools/reporter
    :param hgvs_text: ( c.DNA | r.RNA | p.Protein | g.Genomic )
    :return: JSON (dictionary)
    """
    #r = requests.post("http://www.ncbi.nlm.nih.gov/projects/SNP/VariantAnalyzer/var_rep.cgi", data={"annot1": hgvs_text})
    hgvs_text = str(hgvs_text)
    r = requests.get("http://www.ncbi.nlm.nih.gov/projects/SNP/VariantAnalyzer/var_rep.cgi?annot1={}".format(urllib.parse.quote(hgvs_text)))
    res = r.text

    if 'Error' in res:
        error_str = 'An error occurred when using the NCBI Variant Report Service: "{}"\n'.format(res)
        error_str += 'To reproduce, visit: http://www.ncbi.nlm.nih.gov/projects/SNP/VariantAnalyzer/var_rep.cgi?annot1={}'.format(hgvs_text)
        raise RuntimeError(error_str)

    if IS_DEBUG_ENABLED:
        log.debug(res)

    res = res.split('\n')
    res = filter(
        lambda x: x != '' and not str.startswith(str(x), '.') and not str.startswith(str(x), '##') and not str.startswith(str(x), "Submitted"),
        res)
    res = map(lambda x: x.split('\t'), res)
    keys = map(lambda x: x.strip('# '), res[0])
    values = res[1:]
    res = map(lambda x: dict(zip(keys, x)), values)
    for r in res:
        if r.has_key('PMIDs'):
            if len(r['PMIDs']) == 0:
                r['PMIDs'] = []
            else:
                r['PMIDs'] = r.get('PMIDs').replace(', ', ';').split(';')

    return res


def _ncbi_variant_pubmeds(hgvs_text=None):
    """
    Retrieve PMIDs for a variant from the NCBI Variant Reporter Service
    :param hgvs_text:  ( c.DNA | r.RNA | p.Protein | g.Genomic )
    :return: list(PMID)
    """
    _report  = _ncbi_variant_report_service(hgvs_text)
    _pubmeds = None

    for _row in _report:
        _pubmeds = _row['PMIDs']
        if _pubmeds is not None:
            for _pmid in _pubmeds:
                if len(str(_pmid)) > 1:
                    pass

    return map(int, _pubmeds)

def _service_report_accession(hgvs_text=None):
    """
    Retrieve accession for a variant from the NCBI Variant Reporter Service
    :param hgvs_text: ( c.DNA | r.RNA | p.Protein | g.Genomic )
    :return: Accession (clinvar | dbSNP | dbVar)
    """
    _report  = _ncbi_variant_report_service(hgvs_text)

    for _row in _report:
        _accession = _row['ClinVar Accession']

        if _accession is not None:
            if len(str(_accession)) > 1:
                    return _accession

    log.debug("Variant accession not found in clinvar for "+ str(hgvs_text))
    return None

def _clinvar_variant_accession(hgvs_text):
    """
    See ClinVar FAQ http://www.ncbi.nlm.nih.gov/clinvar/docs/faq/#accs
    :param hgvs_text: c.DNA
    :return: RCVAccession "Reference ClinVar Accession"
    """
    try:
        return ClinVarDB().accession_for_hgvs_text(str(hgvs_text))
    except Exception as err:
        log.debug("no clinvar accession for variant hgvs_text %s " % hgvs_text)

def _clinvar_variant_allele_id(hgvs_text):
    """
    Get the unique AlleleID
    :param hgvs_text: c.DNA
    :return: AlleleID
    """
    try:
        return ClinVarDB().allele_id_for_hgvs_text(hgvs_text)
    except Exception as err:
        log.debug('no clinvar AlleleID for variant hgvs_text %s ' % hgvs_text)

def _clinvar_variant_variation_id(hgvs_text):
    """
    Get the unique VariationID
    :param hgvs_text: c.DNA
    :return: VariationID
    """
    try:
        return ClinVarDB().variation_id_for_hgvs_text(hgvs_text)
    except Exception as err:
        log.debug('no clinvar VariationID for variant hgvs_text %s ' % hgvs_text)

def _clinvar_variant2pubmed(hgvs_text):
    """
    Get PMID for clinvar variants using the AlleleID key.

    Keep GeneReviews book references (NKBxxxx) without argument.

    ONE EXPENSIVE LOOKUP HERE:
    If the citation_source is PubMedCentral, first convert responses to PMID.

    :param hgvs_text: c.DNA
    :return: set(PMIDs and possibly also NBK ids)
    """
    pubmeds = []
    citations = ClinVarDB().var_citations(hgvs_text)
    if citations:
        for cite in citations:
            some_id = cite['citation_id']

            if is_ncbi_bookID(some_id):
                # Todo: convert? drop??
                pubmeds.append(some_id)
            elif is_pmcid(some_id):
                try:
                    pmid = get_pmid_for_otherid(some_id)
                    if pmid is not None:
                        log.debug('found PubMedCentral PMCID %s, converted to PMID %s ', some_id, str(pmid))
                        pubmeds.append(pmid)
                    else:
                        log.debug('PMID not found for PMCID %s; discarding.', some_id)
                except Exception as err:
                    log.debug('error converting PMCID %s: %r', some_id, err)
            elif cite['citation_source'] == 'PubMed':
                pubmeds.append(some_id)

    #return set([int(entry) for entry in pubmeds])
    return set(pubmeds)


def clinvar2pmid_with_accessions(hgvs_list):
    ret = []
    citations = ClinVarDB().var_citations(hgvs_list)
    if citations:
        for cite in citations:
            article_id = cite['citation_id']
            if is_ncbi_bookID(article_id):
                pmid = article_id
            else:
                pmid = article_id if cite['citation_source'] == 'PubMed' else get_pmid_for_otherid(article_id)
            if pmid:
                ret.append({"hgvs_text": cite['HGVS'], "pmid": pmid, "accession": cite['RCVaccession']})
    return ret


##########################################################################################
#
# API
#
##########################################################################################

NCBIVariantReport  = _ncbi_variant_report_service
NCBIVariantPubmeds = _ncbi_variant_pubmeds
ClinvarAccession   = _clinvar_variant_accession
ClinvarAlleleID    = _clinvar_variant_allele_id
ClinvarPubmeds     = _clinvar_variant2pubmed
ClinvarVariationID = _clinvar_variant_variation_id



from .dataset import SQLData

##########################################################################################
#
#       SQLData Class
#
##########################################################################################

class GeneBorg:
    __shared_state = {}
    __gene2pubmed = {}
    __gene2id = {}
    cnt = 0

    def __init__(self, db):
        self.__dict__ = self.__shared_state
        self._db = db

    def gene2pubmed(self, ncbi_gene_id):
        """
        mysql> desc gene.gene2pubmed;
        +--------+------------------+------+-----+---------+-------+
        | tax_id | int(5) unsigned  | YES  | MUL | NULL    |       |
        | GeneID | int(10) unsigned | YES  | MUL | NULL    |       |
        | PMID   | varchar(10)      | YES  | MUL | NULL    |       |
        +--------+------------------+------+-----+---------+-------+
        """
        ncbi_gene_id = self._db.get_gene_id(ncbi_gene_id)
        if ncbi_gene_id not in self.__gene2pubmed:
            self.__gene2pubmed[ncbi_gene_id] = self._db.fetchall('select PMID from gene2pubmed where GeneID = "{}"'.format(ncbi_gene_id))
            self.cnt += 1
        return self.__gene2pubmed[ncbi_gene_id]

    def get_gene_id_for_gene_name(self, hgnc_gene_name_symbol):
        if hgnc_gene_name_symbol not in self.__gene2id:
            self.__gene2id[hgnc_gene_name_symbol] = self._db.fetchID('select GeneID as ID from gene_info where Symbol = "{}" limit 1'.format(hgnc_gene_name_symbol.upper()))
            self.cnt += 1

        return self.__gene2id[hgnc_gene_name_symbol]

class GeneDB(SQLData):
    """
    NCBI Entrez Gene contains links to pubmed (gene2pubmed), MedGen, OMIM, and other sources.
    """
    def __init__(self):
        super(GeneDB, self).__init__(config_section='gene')

    def gene2pubmed(self, ncbi_gene_id):
        """
        Get pubmed entries for gene

        cached: gene2pubmed entries for Entrez GeneID

        :param ncbi_gene_id: int
        :return: list of PMIDs
        """
        return GeneBorg(self).gene2pubmed(ncbi_gene_id)

    def get_gene_id_for_gene_name(self, hgnc_gene_name_symbol):
        """
        cached: Get GeneName (hugo hgnc) for GeneID (ncbi entrez)
        """
        return GeneBorg(self).get_gene_id_for_gene_name(hgnc_gene_name_symbol)

    def gene2mim(self, ncbi_gene_id):
        """
        Online Mendelian Inheritance in Man (OMIM) is a standard reference

        mysql> desc mim2gene_medgen;
        +-----------+------------------+------+-----+---------+-------+
        | Field     | Type             | Null | Key | Default | Extra |
        +-----------+------------------+------+-----+---------+-------+
        | MIM       | int(10) unsigned | YES  | MUL | NULL    |       |
        | GeneID    | int(10) unsigned | YES  | MUL | NULL    |       |
        | MIM_type  | varchar(20)      | YES  |     | NULL    |       |
        | MIM_vocab | varchar(20)      | YES  |     | NULL    |       |
        | MedGenCUI | varchar(20)      | YES  | MUL | NULL    |       |
        +-----------+------------------+------+-----+---------+-------+

        :param ncbi_gene_id: int
        :return: OMIM identifiers with links to MedGen.
        """
        ncbi_gene_id = self.get_gene_id(ncbi_gene_id)
        return self.fetchall('select * from mim2gene_medgen where GeneID = {}'.format(ncbi_gene_id))

    def gene_function(self, ncbi_gene_id):
        """
        get gene-reference-in-function (RIF) for a given gene id.
        This is a textual description of what the gene actually DOES.

        :param ncbi_gene_id:  int
        :return: SQL result GeneRIF with list of pubmeds
        """
        ncbi_gene_id = self.get_gene_id(ncbi_gene_id)
        return self.fetchall('select distinct pubmeds, GeneRIF from generifs_basic where GeneID = {}'.format(ncbi_gene_id))

    def get_gene_id(self, gene):
        """
        Entrez Gene ID.
        (Guard) Ensures we are referring to the NCBI GeneID and not a HGNC gene name
        :param gene: either Entrez gene id or HGNC gene name
        :return: GeneID from NCBI Entrez
        """
        try:
            int(gene)
        except ValueError:
            return self.get_gene_id_for_gene_name(gene)
        except AttributeError:
            return gene.GeneID
        else:
            return gene

    def get_gene_name(self, ncbi_gene_id):
        """
        Get HUGO Gene Name (Symbol) for Entrez gene ID
        :param ncbi_gene_id: int
        :return: string
        """
        ncbi_gene_id = self.get_gene_id(ncbi_gene_id)
        return self.fetchID('select Symbol as ID from gene_info where GeneID = {} limit 1'.format(ncbi_gene_id))

    def get_gene_info(self, ncbi_gene_id):
        """
        NCBI Gene Info for a given gene
        :param ncbi_gene_id: gene id (integer)
        :return: SQL result

        mysql> desc gene.get_gene_info;
        +--------------+------------------+------+-----+---------+-------+
        | Field        | Type             | Null | Key | Default | Extra |
        +--------------+------------------+------+-----+---------+-------+
        | tax_id       | int(5) unsigned  | YES  | MUL | NULL    |       |
        | GeneID       | int(10) unsigned | YES  | MUL | NULL    |       |
        | Symbol       | varchar(50)      | YES  | MUL | NULL    |       |
        | LocusTag     | varchar(20)      | YES  |     | NULL    |       |
        | Synonyms     | text             | YES  |     | NULL    |       |
        | dbXrefs      | text             | YES  |     | NULL    |       |
        | chromosome   | varchar(50)      | YES  |     | NULL    |       |
        | map_loc      | varchar(20)      | YES  |     | NULL    |       |
        | GeneDesc     | text             | YES  |     | NULL    |       |
        | GeneType     | varchar(20)      | YES  |     | NULL    |       |
        | Nomen_symbol | varchar(20)      | YES  |     | NULL    |       |
        | Nomen_source | varchar(20)      | YES  |     | NULL    |       |
        | Nomen_status | varchar(20)      | YES  |     | NULL    |       |
        | GeneOther    | text             | YES  |     | NULL    |       |
        | LastModified | varchar(10)      | YES  |     | NULL    |       |
        +--------------+------------------+------+-----+---------+-------+
        """
        ncbi_gene_id = self.get_gene_id(ncbi_gene_id)
        return self.fetchrow('select * from gene_info where GeneID = %s limit 1'.format(ncbi_gene_id))

    def get_gene_synonyms(self, symbol):
        """
        Get gene synonyms, including symbols that may not be officially recognized.

        :param symbol: Hugo gene name
        :return: list [Synonyms, Symbol, GeneID]
        """
        permutations = (symbol, symbol,       # identity
                        '%|' + symbol,        # symbol at end of list
                        symbol + '|%',        # symbol at start of list
                        '%|' + symbol + '|%', # symbol in middle of list
                        symbol                # identity Nomen_symbol placement
                       )
        _sql = """select Synonyms, Symbol, GeneID from gene_info where 
            Symbol = %s OR
            (Synonyms   = %s or
            Synonyms like %s or
            Synonyms like %s or
            Synonyms like %s) OR
        Nomen_symbol = %s
        """
        return self.fetchall(_sql, *permutations)

    def get_gene_list_from_mim(self, mim):
        mim = str(mim)
        sql = """
            SELECT Symbol AS gene_name FROM gene_info WHERE GeneID IN
            (SELECT  DISTINCT GeneID from mim2gene_medgen WHERE MIM = %s""" 
        return self.list_genes(sql, mim)




from ..db.gene import GeneDB
from ..vocab import Vocab, NCBI_GeneID, HGNC_GeneName

###########################################################################
#
#       Type
#
###########################################################################

class Gene:
    def __init__(self, gene):
        """
         Gene defines both the Entrez GeneID and Hugo GeneName.
        :param gene: gene_id (GeneID) or hgnc (GeneName)
        :return:
        """
        self.id = None
        self.name = None

        # parse gene arg as gene_id (int) or hgnc (str)
        try:
           self.id = int(gene)
           self.name = GeneDB().get_gene_name(gene)

        except ValueError:
            self.name = gene
            self.id = GeneDB().get_gene_id_for_gene_name(gene)

    def __eq__(self, other):
        return (self.__dict__ == other.__dict__)

    def __str__(self):
        return str([
            str(Vocab(NCBI_GeneID, str(self.id))),
            str(Vocab(HGNC_GeneName, self.name))
        ])



##########################################################################################
from collections import OrderedDict

from ..log        import log
from ..db.gene    import GeneDB
from ..db.hugo    import HugoDB
from ..db.clinvar import ClinVarDB

##########################################################################################
#
#       Functions
#
##########################################################################################

def _gene_locus_databases(gene):
    """
    Get Locus Specific Databases for Gene
    :param gene: hugo gene name
    :return: array of LSDBs from Hugo Gene, the official source of gene names.
    """
    try:
        row = HugoDB().get_locus_specific_databases(gene)

        parsed = OrderedDict()
        parsed['Gene'] = str(gene)

        dbs  = OrderedDict()
        text = row['LocusSpecificDatabases']

        if text is not None and len(text) > 1:
            for csv in text.split(','):
                (name, url) = csv.split('|')
                dbs[name] = url

        parsed['LocusSpecificDatabases'] = dbs

        pubmds = []
        text = row['pubmeds']

        if text is not None and len(text) > 0:
            for pmid in text.split(','):
                pmid = pmid.strip(' ')
                pubmds.append(str(pmid)) # @TODO: MetaPub?
                #pubmds.append(PubMed(pmid)) # @TODO: MetaPub?

            parsed['pubmeds'] = pubmds

        return parsed
    except Exception as err:
        msg = 'Failed to fetch LocusSpecificDatabases for gene_symbol'
        log.error(msg)
        raise err

def _gene_synonyms(gene):
    """
    Each gene can have multiple GeneSynonyms -- names that refer to the same gene.
    :param gene: string hugo gene name (hgnc)
    :return: set of gene symbols and names, including unofficial ones.
    """
    synonyms = set()

    if gene is None:
       log.warn('Could not get HGNC GeneName synonyms' )
       return None
    else:
       aliases = GeneDB().get_gene_synonyms(gene)

       if (aliases is None) or (len(aliases) < 1):
           raise Exception('could not retrieve gene SYNONYMS for '+gene)
       else:
           for row in aliases:
               synonyms.add(row.get('Symbol'))

               for gene_alias in row.get('Synonyms').split('|'):
                   if gene_alias.strip():
                       synonyms.add(gene_alias)

    return sorted(synonyms)

def _gene_preferred(gene):
    """
     Each gene can have multiple GeneSynonyms.
     This function returns the one preferred by HUGO.
    :param gene: string gene symbol
    :return: string of preferred symbol
    """
    prefered = set()

    if gene is None:
       return None
    else:
       aliases = GeneDB().get_gene_synonyms(gene)

       if (aliases is None) or (len(aliases) < 1):
           raise Exception('could not retrieve PREFERRED gene name for '+gene)
       else:
           for row in aliases:
               prefered.add(row.get('Symbol'))

    return sorted(prefered)

##########################################################################################
#
#       API
#
##########################################################################################

Gene2PubMed      = GeneDB().gene2pubmed
Gene2Function    = GeneDB().gene_function
Gene2LocusDB     = _gene_locus_databases

Gene2MIM                  = GeneDB().gene2mim
Gene2ConditionSource      = ClinVarDB().gene2condition
Gene2ClinicalSignificance = ClinVarDB().gene_to_clinical_significance_type_frequency

GeneInfo          = GeneDB().get_gene_info
GeneID            = GeneDB().get_gene_id
GeneName          = GeneDB().get_gene_name
GeneSynonyms      = _gene_synonyms
GeneNamePreferred = _gene_preferred





from .dataset import SQLData

################################################################################
#
# Formatting helper functions
#
################################################################################

def is_format_umls(concept='C0007194'):
    """
    CUI style notation
    :param concept:
    :return: boolean
    """
    return str(concept).startswith('C') and (8)== str(concept).__len__()

def is_format_medgen(concept=2881):
    """
    UID style notation
    :param concept:
    :return: boolean
    """
    try:
        concept = int(concept)
        return str(concept).__len__() <= (6)
    except ValueError:
        return False

##########################################################################################
#
#       SQLData Class
#
##########################################################################################

class MedGenDB(SQLData):
    """
    MedGen is the NCBI primary source for Medical Genetics.
    http://www.ncbi.nlm.nih.gov/books/NBK159970/
    https://bitbucket.org/invitae/medgen-mysql
    """

    def __init__(self):
        super(MedGenDB, self).__init__(config_section='medgen')

    def disease_subtypes(self, cui):
        """
        Narrower Hierarchical Relationship:
        Disease subtypes refers to an IS-A relationship in the medgen term hierarchy.
        :param cui: MedGen concept ID
        :return: id, name, and source of the disease concept
        """
        cui = self.get_concept_id(cui)
        select_template = '''
        select distinct
        SubtypeID     as DiseaseID,
        SubtypeName   as DiseaseName,
        SubtypeSource as DiseaseSource
        from view_disease_subtype where DiseaseID="%s";'''
        return self.fetchall(select_template % cui)

    def disease_parents(self, cui):
        """
        Broader Hierarchical Relationship:
        Disease parents refers to more general definitions of the disease specified by the cui param.
        This table is created by invitae, not downloaded directly from medgen.
        :param cui: MedGen concept
        :return: id, name, and source of the disease
        """
        cui = self.get_concept_id(cui)
        select_template = '''select distinct DiseaseID, DiseaseName, DiseaseSource from view_disease_subtype where SubTypeID="%s";'''
        return self.fetchall(select_template % cui)

    def concept_name(self, cui):
        """
        Get preferred concept name, NCBI MedGen first prefers GTR/ClinVar concept names, then SNOMED-CT, followed by MESH.
        Most concept names comes from SNOMED-CT (Clinical Terms)

        Example entry:
        +----------+--------------------------------------------------------+-------------+----------+
        | CUI      | name                                                   | source      | SUPPRESS |
        +----------+--------------------------------------------------------+-------------+----------+
        | C1997451 | History of repair of atrial septal defect              | SNOMEDCT_US | N        |
        +----------+--------------------------------------------------------+-------------+----------+

        :param cui: medgen concept
        :return: conept name
        """
        return self.fetchrow("select * from NAMES where CUI = '?' "
                             .replace('?', str(self.get_concept_id(cui))))


    def concept_definition(self, cui):
        """
        Get concept definition (if available)
        SAB refers to "source vocabulary"

        Example entry:
        +----------+------------------------------------------------------------------+-----+----------+
        | CUI      | DEF                                                              | SAB | SUPPRESS |
        +----------+------------------------------------------------------------------+-----+----------+
        | C0266139 | Congenital tracheoesophageal fistula without esophageal atresia. | NCI | N        |
        +----------+------------------------------------------------------------------+-----+----------+

        :param cui: medgen concept
        :return: dict(CUI, DEF, SAB)
        """
        return self.fetchrow("select * from MGDEF where CUI = '?' "
                             .replace('?', str(self.get_concept_id(cui))))


    def concept_relations(self, cui):
        """
        Relate concepts.
        Relationships were sources from UMLS, the Unified Medical Language System.

        Each relationship is defined by 'REL' and optionally

        http://www.ncbi.nlm.nih.gov/books/NBK9685/
        http://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/columns_data_elements.html
        http://www.nlm.nih.gov/research/umls/implementation_resources/query_diagrams/er1.html

        Counts of each type:
        select REL,count(*) as cnt from MGREL group by REL order by cnt desc
        +-----+--------+
        | REL | cnt    |
        +-----+--------+
        | RO  | 735308 | Relate "Other"
        | RB  | 138570 | Relate "Broader"
        | RN  | 138570 | Relate "Narrower"
        | SY  | 112964 | Relate "Synonym"
        | SIB |  92934 | Sibling.
        | CHD |  85812 | Child.
        | PAR |  85812 | Parent.
        | AQ  |  21388 |
        | QB  |  21388 |
        | RQ  |  18946 |
        +-----+--------+

        +-----+--------------------------------------------------------+--------+
        | REL | RELA                                                   | cnt    |
        +-----+--------------------------------------------------------+--------+
        | RO  | has_manifestation                                      | 177270 |
        | RO  | manifestation_of                                       | 177270 |
        | RN  | mapped_to                                              |  93163 |
        | RB  | mapped_from                                            |  93163 |
        | SIB |                                                        |  92934 |
        | RO  | disease_has_finding                                    |  49583 |
        | RO  | is_finding_of_disease                                  |  49583 |
        | PAR | inverse_isa                                            |  47643 |
        | CHD | isa                                                    |  47643 |
        | SY  | permuted_term_of                                       |  39654 |
        | SY  | has_permuted_term                                      |  39654 |
        ...


        mysql> desc MGREL;
        +----------+--------------+------+-----+---------+-------+
        | Field    | Type         | Null | Key | Default | Extra |
        +----------+--------------+------+-----+---------+-------+
        | CUI1     | char(8)      | NO   | MUL | NULL    |       |
        | AUI1     | varchar(9)   | YES  |     | NULL    |       |
        | STYPE1   | varchar(50)  | NO   |     | NULL    |       |
        | REL      | varchar(4)   | NO   | MUL | NULL    |       |
        | CUI2     | char(8)      | NO   | MUL | NULL    |       |
        | AUI2     | varchar(9)   | YES  |     | NULL    |       |
        | RELA     | varchar(100) | YES  | MUL | NULL    |       |
        | RUI      | varchar(10)  | NO   |     | NULL    |       |
        | SAB      | varchar(40)  | NO   |     | NULL    |       |
        | SL       | varchar(40)  | NO   |     | NULL    |       |
        | SUPPRESS | char(1)      | NO   |     | NULL    |       |
        +----------+--------------+------+-----+---------+-------+

        :param cui: concept id
        :return: relationships defined in MGREL table
        """
        return self.fetchall("select * from MGREL where CUI1 = '?' or CUI2 = '?' "
                             .replace('?', str(self.get_concept_id(cui))))

    def concept_sources(self, cui):
        """
        Get source vocabulary names (dictionary abbreviations) for a given concept.
        :param cui: MedGen concept
        :return: list of dictionary names (SourceVocab)
        """
        return self.fetchall("select distinct SourceVocab from view_concept where ConceptID = '?' "
                             .replace('?', str(self.get_concept_id(cui))))

    def medgen2umls(self, medgen_uid):
        """
        Get CUI for UID. This is for compatibility purposes.

        :param medgen_uid: int like 651, which points to C0006142
        :return: concept code like "C0006142"
        """
        select_template = '''select ConceptID as ID from view_medgen_uid where MedGenUID = "%s";'''
        return self.fetchID(select_template % medgen_uid)

    def umls2medgen(self, cui):
        """
        Get UID for concept id. This is for compatibility purposes.

        :param cui: concept code like "C0006142"
        :return: int like 651, which points to C0006142
        """
        select_template = '''select MedGenUID as ID from view_medgen_uid where ConceptID = "%s";'''
        return self.fetchID(select_template % cui)

    def get_concept_id(self, unique_id):
        """
        Guard: force usage of MedGen CUI based identifier.

        :param unique_id: either CUI or a UID
        :return: CUI
        """
        if is_format_umls(unique_id):
            return unique_id

        if is_format_medgen(unique_id):
            return self.medgen2umls(unique_id)

        raise Exception('Unknown concept unique identifier format for %s' + unique_id)

    def select_hpo_view_medgen_hpo(self, cui):
        """
        HPO Human Phenotype Ontology
        :param cui: MedGen concept
        :return: MySQL result
        +---------------+--------------+------+-----+---------+-------+
        | Field         | Type         | Null | Key | Default | Extra |
        +---------------+--------------+------+-----+---------+-------+
        | SourceVocab   | varchar(20)  | NO   | MUL | NULL    |       |
        | SemanticType  | varchar(50)  | NO   | MUL | NULL    |       |
        | ConceptID     | char(8)      | NO   | MUL | NULL    |       |
        | ConceptName   | text         | NO   |     | NULL    |       |
        | PhenotypeName | text         | NO   |     | NULL    |       |
        | HPO_ID        | varchar(100) | YES  | MUL | NULL    |       |
        +---------------+--------------+------+-----+---------+-------+
        """
        return self.fetchall(
            "select * from medgen.view_medgen_hpo where ConceptID = '%s' " % cui)

    def select_mim_from_cui(self, cui):
        return self.fetchlist("select DISTINCT MIM_number from medgen_hpo_omim where omim_cui='{}'".format(cui), "MIM_number")





from .dataset import SQLData
from ..parse.concept import Concept
from ..parse.gene import Gene
from ..log import log, IS_DEBUG_ENABLED

##########################################################################################
#
#       SQLData Class
#
##########################################################################################

class PersonalGenomesDB(SQLData):
    """
    PGP contains annotations using BioNotate.
    Contributors welcome!
    """

    def __init__(self):
        super(PersonalGenomesDB, self).__init__(config_section='personalgenomes')

    def bionotate__gene_aa_pos(self, hgnc, variant_aa_pos):
        """
        human curated variants from bionotate / Personal Genomes Project
        :param hgnc: string hugo gene name
        :param variant_aa_pos: amino acid position
        :return:
        """
        _sql = """
        SELECT distinct
          PMID, 
          variant_gene as gene,
          variant_id   as vid,
          variant_aa_del,
          variant_aa_pos ,
          variant_aa_ins
        FROM   bionotate
        WHERE  gene = '?gene' and variant_aa_pos = '?pos'
        """
        return self.fetchall(_sql.replace('?gene', hgnc).replace('?pos', variant_aa_pos))





import sys
import logging
import datetime
from  collections import OrderedDict

log = logging.getLogger('medgen')

##########################################################################
# LOGGING
#
#

HGVS_LOG_LEVEL = logging.INFO

#console = logging.StreamHandler(sys.stdout)
#console.setFormatter(logging.Formatter('%(module)s  %(funcName)s(%(lineno)d): %(levelname)s %(message)s'))

log.setLevel(HGVS_LOG_LEVEL)
#log.addHandler(console)

IS_DEBUG_ENABLED = False #log.isEnabledFor(logging.DEBUG)
IS_INFO_ENABLED  = log.isEnabledFor(logging.INFO)

biomed_timer_start = datetime.datetime.now()
biomed_timer_stop  = datetime.datetime.now()

class LogTimer:

    def __init__(self):
        self._start_time = None
        self._stop_time  = None
        self.passed = OrderedDict()
        self.failed = OrderedDict()
        self.restart()

    def restart(self):
        "Start or restart"
        self._start_time= datetime.datetime.now()
        self._stop_time = None
        log.debug(str(self))

    def log_passed(self, item, result=None):
        log.debug(str(item)+':'+str(result))
        self.passed[item] = result

    def log_failed(self, item, error=None):
        log.warn(str(item)+':'+str(error))
        self.failed[item] = error

    def elapsed(self):
        self._stop_time = datetime.datetime.now()

        diff = (self._stop_time - self._start_time)

        log.info('elapsed:'+ str(diff))
        log.info('passed:'+ str(len(self.passed.keys())))
        log.info('failed:'+ str(len(self.failed.keys())))

        return diff

    def __str__(self):
        return '@LogTimer '+ str(self.__dict__)


def LogContainer(_func, _collection):
    t = LogTimer()
    for item in _collection:
        try:
            t.log_passed(item, _func(item))

        except Exception as e:
            t.log_failed(item, e)
            t.elapsed()

    t.elapsed()
    return t




from .dataset import SQLData
from ..parse.concept import Concept
from ..parse.gene import Gene
from ..log import log, IS_DEBUG_ENABLED

##########################################################################################
#
#       SQLData Class
#
##########################################################################################

class ClinVarDB(SQLData):
    """
    ClinVar is the NCBI database of clinical variants.
    ClinVar links MedGen phenotypes to HGVS variants and PubMed articles
    """
    def __init__(self):
        super(ClinVarDB, self).__init__(config_section='clinvar')

    def clinvar_ids(self, hgvs_text, id_column='VariationID'):
        """
        Get clinvar identifier for HGVS text label.
        ClinVar XML contains all of the recognizable HGVS strings.
        See medgen-mysql for specifics on how HGVS and ClinVar identifiers are associated.

        :param hgvs_text: c.DNA, r.RNA, p.Protein, or g.Genomic
        :param id_column: 'VariationID', 'AlleleID', or 'RCVaccession'
        :return: clinvar identifer 'VariationID', 'AlleleID', or 'RCVaccession'
        """
        res = self.fetchall(
            ' select distinct %s as ID ' % id_column +
            ' from clinvar_hgvs ' +
            ' where HGVS = "%s"' % hgvs_text)

        return [entry['ID'] for entry in res]

    def accession_for_hgvs_text(self, hgvs_text):
        """
        Get RCVaccession for hgvs_text
        :param hgvs_text: c.DNA, r.RNA, p.Protein, or g.Genomic
        :return: RCVaccession
        """
        return self.clinvar_ids(hgvs_text, 'RCVaccession')

    def allele_id_for_hgvs_text(self, hgvs_text):
        """
        Get AlleleID for hgvs_text
        :param hgvs_text: c.DNA, r.RNA, p.Protein, or g.Genomic
        :return: AlleleID
        """
        return self.clinvar_ids(hgvs_text, 'AlleleID')

    def variation_id_for_hgvs_text(self, hgvs_text):
        """
        Get VariationID, the NCBI preferred identifier in ClinVar.
        :param hgvs_text: c.DNA, r.RNA, p.Protein, or g.Genomic
        :return: VariationID
        """
        return self.clinvar_ids(hgvs_text, 'VariationID')

    def hgvs_text_for_variation_id(self, variation_id):
        """
        Get hgvs_text for the specified NCBI Variant
        :param variation_id: Identifier preferred by NCBI ClinVar for a given Variant
        :return: arry of hgvs_text like ['NM_000530.6:c.233C>A']
        """
        res = self.fetchall("select distinct HGVS as ID from clinvar_hgvs where VariationID = %s " % str(variation_id))

        return [entry['ID'] for entry in res]

    def variant_summary(self, hgvs_c, hgvs_r=None, hgvs_p=None):
        """
        Get summary of a variant using the HGVS text label.
        The NCBI preferred identifier is VariationID, so use that instead if possible.

        To get a list of VariantSynonyms to work with this function, see the hgvs-lexicon package.
        hgvs-lexicon Maps HGVS synonyms names. We search with these synonyms where possible.

        [ hgvs_c, hgvs_r, hgvs_p] = VariantSynonyms(hgvs_c)

       +----------------------+--------------+------+-----+---------+-------+
       | Field                | Type         | Null | Key | Default | Extra | Invitae
       +----------------------+--------------+------+-----+---------+-------+
       | AlleleID             | int(11)      | NO   | MUL | NULL    |       | We use AlleleID instead of accession(s) to attempt mapping uniquely to clinvar
       | variant_type         | varchar(50)  | NO   | MUL | NULL    |       |
       | variant_name         | text         | YES  |     | NULL    |       |
       | GeneID               | int(11)      | NO   | MUL | NULL    |       | medgen.parse.gene.GeneID
       | Symbol               | varchar(20)  | NO   | MUL | NULL    |       | medgen.parse.gene.GeneName
       | ClinicalSignificance | text         | YES  |     | NULL    |       |
       | rs                   | int(11)      | YES  | MUL | NULL    |       |
       | dbvar_nsv            | text         | YES  |     | NULL    |       |
       | RCVaccession         | text         | YES  |     | NULL    |       |
       | TestedInGTR          | char(1)      | YES  | MUL | NULL    |       | GTR  !!!
       | PhenotypeIDs         | text         | YES  |     | NULL    |       | @todo: parse
       | Origin               | text         | YES  |     | NULL    |       |
       | Assembly             | text         | YES  |     | NULL    |       |
       | Chromosome           | varchar(20)  | YES  |     | NULL    |       |
       | Start                | int(11)      | YES  |     | NULL    |       |
       | Stop                 | int(11)      | YES  |     | NULL    |       |
       | Cytogenetic          | text         | YES  |     | NULL    |       |
       | ReviewStatus         | text         | YES  |     | NULL    |       |
       | HGVS_c               | varchar(200) | YES  | MUL | NULL    |       | hgvs_lexicon.parse.Variant
       | HGVS_p               | varchar(200) | YES  | MUL | NULL    |       | hgvs_lexicon.parse.Variant
       | NumberSubmitters     | int(11)      | YES  |     | NULL    |       |
       | LastEvaluated        | text         | YES  |     | NULL    |       |
       | Guidelines           | text         | YES  |     | NULL    |       |
       | OtherIDs             | text         | YES  |     | NULL    |       |
       +----------------------+--------------+------+-----+---------+-------+

        :param hgvs_c: c.DNA variant
        :param hgvs_r: r.RNA variant
        :param hgvs_p: p.Protein variant
        :return: variant summary as described in the clinvar TSV download file.
        """
        if IS_DEBUG_ENABLED:
            log.debug('hgvs_c '+str(hgvs_c))
            log.debug('hgvs_r '+str(hgvs_r))
            log.debug('hgvs_p '+str(hgvs_p))

        _select = """
             Select distinct TestedInGTR,
                  HGVS_c, HGVS_p, variant_name,
                  GeneID, Symbol,
                  variant_type, ClinicalSignificance, PhenotypeIDs,
                  AlleleID, dbvar_nsv, rs,
                  NumberSubmitters """

        _from   = """ From variant_summary """

        _where  = """ Where   (HGVS_c = '@c') or
                              (HGVS_c = '@r') or
                              (HGVS_p = '@p') """

        _where = _where.\
            replace('@c', str(hgvs_c)).\
            replace('@r', str(hgvs_r)).\
            replace('@p', str(hgvs_p))

        _sql = ' '+ _select + _from + _where

        if IS_DEBUG_ENABLED:
            log.debug(str( _sql   ))

        return self.fetchall(_sql)

    def var_citations(self, hgvs_text):
        """
        Get citations for clinvar entries using an HGVS text label.

        +----------+-------------+-----------+------+-----------------+-------------+
        | AlleleID | VariationID | rs        | nsv  | citation_source | citation_id |
        +----------+-------------+-----------+------+-----------------+-------------+
        |    15041 |           2 | 397704705 |    0 | PubMed          |    20613862 |
        |    15041 |           2 | 397704705 |    0 | PubMed          |    20613862 |
        |    15042 |           3 | 397704709 |    0 | PubMed          |    20613862 |
        |    15042 |           3 | 397704709 |    0 | PubMed          |    20613862 |
        |    15043 |           4 | 150829393 |    0 | PubMed          |    12030328 |
        |    15043 |           4 | 150829393 |    0 | PubMed          |    20531441 |
        |    15043 |           4 | 150829393 |    0 | PubMed          |    12030328 |
        |    15043 |           4 | 150829393 |    0 | PubMed          |    20531441 |
        |    15044 |           5 | 267606829 |    0 | PubMed          |    20818383 |
        +----------+-------------+-----------+------+-----------------+-------------+

        :param hgvs_text: c.DNA, r.RNA, p.Protein, g.Genomic
        :return: citations from ClinVar
        """
        if type(hgvs_text) == str:
            hgvs_text = [hgvs_text]

        hgvs_clause = " or ".join(map(lambda x: " H.HGVS = '{}' ".format(x), hgvs_text))
        sql = "select C.citation_id, C.citation_source, H.RCVaccession, H.HGVS from clinvar_hgvs H, var_citations C where H.VariationID = C.VariationID and ({})".format(hgvs_clause)
        print(sql)
        return self.fetchall(sql)


    def molecular_consequences(self, hgvs_text):
        """
         Get Molecular Consequences for hgvs variant.
         Not all clinvar variants have known consequences.
         Sequence Ontology provides standard variant consequence naming.

        +--------------------+--------------+------+-----+---------+-------+
        | Field              | Type         | Null | Key | Default | Extra |
        +--------------------+--------------+------+-----+---------+-------+
        | HGVS               | varchar(200) | YES  | MUL | NULL    |       |
        | SequenceOntologyID | varchar(20)  | YES  | MUL | NULL    |       |
        | Consequence        | varchar(100) | YES  | MUL | NULL    |       |
        +--------------------+--------------+------+-----+---------+-------+

        :param hgvs_text:
        :return:
        """
        return self.fetchall("select * from clinvar.molecular_consequences where HGVS='?'  ".replace("?", hgvs_text))

    def random_example_hgvs(self, num_examples=1):
        """
        Get random mRNA variants in table clinvar.molecular_consequences
        :param num_examples: how many hgvs strings do you want?
        :return: list of hgvs strings

        :param num_examples:
        :return: list of hgvs variants
        """
        cmd = 'select distinct(HGVS) as HGVS from molecular_consequences where HGVS like "NM%" order by rand() limit {}'.format(num_examples)
        return [item['HGVS'] for item in self.fetchall(cmd)]

    def disease_name(self, concept):
        """
        MedGen (including ClinVar) refers to this standard set of disease names.
        :param concept:  UMLS or MedGen concept ID
        :return: DiseaseName entry (dictionary, includes VocabSource)
        """
        return self.fetchall(
         "select * "
         "from disease_names "
         "where ConceptID = '%s'" % Concept(concept).umls_cui)

    #TODO: @nthmost: refactor with medgen-services
    def gene2condition(self, gene_id):
        """
        gene2condition is a MedGen linked source spanning MedGen, ClinVar, GTR, OMIM, and HPO.

        desc gene_condition_source_id
        +--------------+---------------+------+-----+---------+-------+
        | Field        | Type          | Null | Key | Default | Extra |
        +--------------+---------------+------+-----+---------+-------+
        | GeneID       | int(11)       | YES  | MUL | NULL    |       |
        | Symbol       | varchar(10)   | YES  | MUL | NULL    |       |
        | ConceptID    | varchar(20)   | YES  | MUL | NULL    |       |
        | DiseaseName  | varchar(1000) | YES  | MUL | NULL    |       |
        | SourceName   | varchar(150)  | YES  | MUL | NULL    |       |
        | SourceID     | varchar(50)   | YES  | MUL | NULL    |       |
        | DiseaseMIM   | varchar(20)   | YES  | MUL | NULL    |       |
        | LastModified | varchar(20)   | YES  |     | NULL    |       |
        +--------------+---------------+------+-----+---------+-------+

        :param gene_id: Entrez Gene ID
        :return: condition information from gene_condition_source_id
        """
        return self.fetchall(
            " select * from gene_condition_source_id where GeneID = ?".replace('?',str(Gene(gene_id).id)))

    def gene2condition_for_concept(self, concept_id):
        """
        See gene2condition. Input is a concept rather than Gene.
        :param concept_id: MedGen concept id (CUI)
        :return: MedGen linked entry
        """
        return self.fetchall(
            " select * from gene_condition_source_id where ConceptID = '?' ".replace('?',str(concept_id)))


    def gene_summary(self, gene):
        """
        Get ClinVar gene summary

        mysql> select * from gene_specific_summary limit 20 ;
        +-------------+-----------+-------------+---------+
        | Symbol      | GeneID    | Submissions | Alleles |
        +-------------+-----------+-------------+---------+
        | A1BG        |         1 |           6 |       6 |
        | A1CF        |     29974 |           7 |       7 |
        | A2M         |         2 |          14 |      11 |
        | A2ML1       |    144568 |          15 |      14 |
        | A3GALT2     |    127550 |           1 |       1 |
        | A4GALT      |     53947 |          14 |      14 |
        | A4GNT       |     51146 |           5 |       5 |
        | AAAS        |      8086 |          14 |      14 |
        | AACS        |     65985 |           8 |       8 |
        | AADAC       |        13 |           7 |       7 |
        +-------------+-----------+-------------+---------+

        :param gene: NCBI Gene ID or hugo gene name
        :return: dictionary with counts of Submissions and Alleles
        """
        return self.fetchrow("select * from gene_specific_summary where GeneID = ? limit 1".replace('?', str(Gene(gene).id)))


    def gene_to_clinical_significance_type_frequency(self, gene):
        """
        Get clinical significance for a gene.

        ftp://ftp.ncbi.nlm.nih.gov/pub/GTR/standard_terms/Clinical_significance.txt

        Pathogenic
        Likely pathogenic
        drug response
        confers sensitivity
        risk factor
        other
        association
        Uncertain significance
        Likely benign
        Benign
        protective
        not provided
        conflicting data from submitters

        :param gene: NCBI Gene ID or hugo gene name
        :return: dict containing ClinicalSignificance and the number of variants (cnt_variants)
        """
        gene = Gene(gene)
        return self.fetchall(" select ClinicalSignificance, count(*) as cnt_variants "
                             " from variant_summary "
                             " where GeneID = ? "
                             " group by ClinicalSignificance "
                             " order by cnt_variants    desc "
                             .replace('?', str(gene.id)) )


    # TODO: deprecated
    def select_clinvar_gene_list_for_cui(self, cui):
        """
        select ClinVar genes associated with MedGen CUI
        :param cui:
        :return:
        """
        return self.list_genes("select Symbol as gene_name from gene_condition_source_id where ConceptID = '%s' " % cui)

    def get_version(self):
        '''
        Get the ClinVar version
        :return: string with version, such as '2015-10-2'
        '''
        try:
            sql_query = "select version from version_info order by last_loaded desc limit 1"
            return self.fetchrow(sql_query)['version']
        except:
            log.debug('No version numbers found for ClinVar')
            return None




from ..db.medgen import MedGenDB
from ..vocab import Vocab

###########################################################################
#
#       Type
#
###########################################################################

class Concept:
    """
    MedGen
    The NCBI Handbook [Internet]. 2nd edition.
    http://www.ncbi.nlm.nih.gov/books/NBK159970/
    """
    def __init__(self, concept):
        """
        MedGen concept
        :param concept: either umls_cui (UMLS unique concept) or medgen_uid (MedGen defined ID)
        :return: concept with both UMLS and MedGen defined IDs available.
        """
        try:
            self.umls_cui  = None
            self.medgen_uid= None

            if concept:
                self.medgen_uid = int(concept)
                self.umls_cui= str(MedGenDB().medgen2umls(concept))

        except ValueError:
            self.umls_cui  = str(concept)
            self.medgen_uid= int(MedGenDB().umls2medgen(concept))

    def __eq__(self, other):
        return (self.__dict__ == other.__dict__)

    def __str__(self):
        return str([
            str(Vocab("UMLS:ConceptID", self.umls_cui)),
            str(Vocab("MedGen:MedGenUID", self.medgen_uid))
        ])



from .log import log
####################################################################################
class MedGenError(Exception):

    def __init__(self, error=None, cause=None):
        self.error = error
        self.cause = cause

    def __str__(self):
        return str(__dict__)

class ConnectError(MedGenError):
    """
    caused by connection errors in
    HGVS Parser, Mapper, UTA, or PubTator.
    Thrown when the annotator cannot load a resource
    such as the HGVS grammar file or sql databases.
    """
    pass

class ParseError(MedGenError):
    def __init__(self, issue=None, cause=None):
        self.issue = issue
        self.cause = cause

    def __str__(self):
        return str(self.__dict__)

####################################################################################

class DBError(MedGenError): pass

class AnnotateError(MedGenError):pass

class TODO(MedGenError): pass

####################################################################################

class JIRA_TODO(MedGenError):
    def __init__(self, issue=None, cause=None):
        if  issue==None:
            issue='Not Yet Implemented'

        self.issue = str('https://invitae.jira.com/browse/'+ str(issue))
        self.cause = cause

    def __str__(self):
        return str(self.__dict__)

####################################################################################
#
# GUARDS (usage optional) 
# 
####################################################################################

def GuardNotNone(echo, strict=True):
    """
    This checks whether an object is None.
    If echo is None and strict is set to True (default), throws an exception
    (or whatever GuardError has now been set to do.
    If echo is None and strict is set to False, will return None.

    :param echo: the variable to be checked
    :param strict: whether an exception should be thrown
    :return: the original object (echo)
    """
    log.debug(echo)

    return \
        echo if (echo is not None) else \
        None if (strict is not True) else GuardError(echo, strict)

def GuardAssert(echo, assertion=False, strict=True):
    return \
        echo if assertion is True else \
        GuardNotNone(echo,strict) if not strict else \
        GuardError(echo, strict, assertion)


def GuardError(echo, assertion=False, strict=True):
    error = ' echo ='+str(echo) + " strict ="+str(strict)+ " assertion="+str(assertion)
    if strict:
        raise Exception('Guard error: found value is None' )#error)
    else:
        print('GuardError(?)' + error)



from .dataset import SQLData
from ..parse.gene import Gene

##########################################################################################
#
#       SQLData Class
#
##########################################################################################

class HugoDB(SQLData):
    """
    HUGO gene names, info, and locus specific databases.
    """

    def __init__(self):
        super(HugoDB, self).__init__(config_section='hugo')

    def hugo_info(self, gene_symbol):
        """
        Get gene information, official records from the standard HUGO gene committee

        mysql> desc hugo_info;
        +------------------------+-------------+------+-----+---------+-------+
        | Field                  | Type        | Null | Key | Default | Extra |
        +------------------------+-------------+------+-----+---------+-------+
        | hgnc                   | varchar(25) | YES  |     | NULL    |       |
        | Symbol                 | varchar(50) | YES  |     | NULL    |       | variant2pubmed.parse.gene
        | Name                   | text        | YES  |     | NULL    |       |
        | Status                 | text        | YES  |     | NULL    |       |
        | LocusType              | text        | YES  |     | NULL    |       |
        | LocusGroup             | text        | YES  |     | NULL    |       |
        | PreviousSymbols        | text        | YES  |     | NULL    |       |
        | PreviousNames          | text        | YES  |     | NULL    |       |
        | Synonyms               | text        | YES  |     | NULL    |       |
        | NameSynonyms           | text        | YES  |     | NULL    |       |
        | Chromosome             | text        | YES  |     | NULL    |       |
        | DateNameChanged        | text        | YES  |     | NULL    |       |
        | AccessionNumbers       | text        | YES  |     | NULL    |       |
        | EnsemblGeneID          | int(10)     | YES  |     | NULL    |       |
        | SpecialistDB           | text        | YES  |     | NULL    |       |
        | pubmeds                | text        | YES  |     | NULL    |       | todo: check: same as gene2pubmed?
        | RefSeqIDs              | text        | YES  |     | NULL    |       |
        | GeneFamilyTag          | text        | YES  |     | NULL    |       |
        | RecordType             | text        | YES  |     | NULL    |       |
        | PrimaryIDs             | text        | YES  |     | NULL    |       |
        | LocusSpecificDatabases | text        | YES  |     | NULL    |       |
        +------------------------+-------------+------+-----+---------+-------+

        :param gene_symbol:
        :return:
        """
        return self.fetchall("select * from hugo.hugo_info where Symbol = '?' ".
                             replace('?', str(gene_symbol)))


    def get_locus_specific_databases(self, gene_symbol):
        """
        get LSDBs for a given gene symbol.

        :param gene_symbol:
        :return: dict {LocusSpecificDatabases, GeneFamilyTag, pubmeds}
        """
        gene = Gene(gene_symbol)

        return self.fetchrow(
            "select LocusSpecificDatabases, GeneFamilyTag, pubmeds "
            "from hugo_info where Symbol = '?' ".
            replace('?', str(gene.name)))



##########################################################################
# parse

from .parse.gene    import Gene
from .parse.concept import Concept

##########################################################################
# db

from .db.dataset import SQLData
from .db.medgen  import MedGenDB
from .db.gene    import GeneDB
from .db.hugo    import HugoDB
from .db.personalgenomes import PersonalGenomesDB
from .db.clinvar import ClinVarDB

##########################################################################
# annotate

from .annotate.variant import ClinvarPubmeds, ClinvarAccession, ClinvarAlleleID, ClinvarVariationID

from .annotate.gene    import GeneID, GeneName
from .annotate.gene    import GeneInfo,    GeneSynonyms, GeneNamePreferred
from .annotate.gene    import Gene2PubMed, Gene2LocusDB, Gene2Function
from .annotate.gene    import Gene2MIM,    Gene2ConditionSource, Gene2ClinicalSignificance

from .annotate.disease import DiseaseName, DiseaseParents, DiseaseSubtypes
from .annotate.concept import ConceptName, ConceptDefinition, ConceptRelations, ConceptSources, Define, Relate

from .annotate.pubmed import PMCID2Article, PMID2Article

##########################################################################



#### metapub
from metapub import PubMedFetcher, PubMedArticle
from medgen.db.medgen import MedGenDB

##########################################################################################
#
#  Functions
#
##########################################################################################

def _pubmed_pmid_to_article(pmid):
    """
    Use NCBI eutils to fetch pubmed article information.

    :param pmid: int or str
    :return: PubMedArticle
    """
    return PubMedFetcher().article_by_pmid(str(pmid))

def _pubmed_central_pmcid_to_article(pmcid):
    """
    Specific to PMC PubMed Central.
    Use NCBI eutils to fetch pubmed article information.

    :param pmcid:
    :return: PubMedArticle
    """
    return PubMedFetcher().article_by_pmcid(str(pmcid))

##########################################################################################
#
#       API
#
##########################################################################################
PMID2Article = _pubmed_pmid_to_article
PMCID2Article = _pubmed_central_pmcid_to_article



import re
import datetime
from .dataset import SQLData

##########################################################################################
#
#       SQLData Class
#
##########################################################################################


class PubMedDB(SQLData):
    """
    Common access methods for PubMed tables.
    See also https://pypi.python.org/pypi/metapub
    """
    def __init__(self):
        super(PubMedDB, self).__init__(config_section='pubmed')

    def abstract_text(self, pmid):
        """
        Get the abstract for a pmid
        :var pmid : pmid
        :returns abstract_text
        """
        sql = """
            select abstract_text
            from   medline_minimum_citation
            where  PMID = '@PMID'
            """. \
            replace("@PMID", str(pmid))
        return self.fetchrow(sql)

    def medline_xml_filename_insert(self, filename):
        """
        Book Keeping method for Medline XML
        :param filename: name of medline_xml file
        :return: result of SQL insert
        """
        return self.insert("medline_xml_filename", {"filename" : filename})

    def _get_tstamp(self, xml_block):
        """
        <DateCreated>
        <Year>2001</Year>
        <Month>07</Month>
        <Day>16</Day>
        </DateCreated>

        :return datetime for day
        """
        y_match = re.search(r"<Year>(\d+)</Year>", xml_block)
        m_match  = re.search(r"<Month>(\d+)</Month>", xml_block)
        d_match  = re.search(r"<Day>(\d+)</Day>", xml_block)
        if y_match and m_match and d_match:
            y = int(y_match.group(1))
            m = int(m_match.group(1))
            d = int(d_match.group(1))
            return datetime.date(year=y, month=m, day=d)
        return None

    def _get_last_date(self, medline_citation_xml_text):
        """
        Get the last revision/completed/or created date:
        <DateCreated>
        <Year>2001</Year>
        <Month>06</Month>
        <Day>21</Day>
        </DateCreated>
        <DateCompleted>
        <Year>2001</Year>
        <Month>07</Month>
        <Day>05</Day>
        </DateCompleted>
        <DateRevised>
        <Year>2010</Year>
        <Month>11</Month>
        <Day>18</Day>
        </DateRevised>
        :param medline_citation_xml_text: "<MedlineCitation>...</MedlineCitation>" xml text
        :return (year, month, date) as integer strings, like ('2013', '09', '15')
        """
        match_revised = re.search(r"<DateRevised>.+</DateRevised>", medline_citation_xml_text, re.DOTALL)
        if match_revised:
            return self._get_tstamp(match_revised.group(0))

        match_completed = re.search(r"<DateCompleted>.+</DateCompleted>", medline_citation_xml_text, re.DOTALL)
        if match_completed:
            return self._get_tstamp(match_completed.group(0))

        match_created = re.search(r"<DateCreated>.+</DateCreated>", medline_citation_xml_text, re.DOTALL)
        if match_created:
            return self._get_tstamp(match_created.group(0))

        raise RuntimeError("Could not find date in %s" % medline_citation_xml_text)

    def _get_PMID(self, medline_citation_xml_text):
        """
        Parse out the PMID from medline XML
        :param medline_citation_xml_text: string containing medline xml
        :return: PMID as a str
        """
        match = re.search(r'<PMID Version="\d+">(\d+)</PMID>', medline_citation_xml_text)
        return match.group(1)

    def medline_xml_select_ids(self, min_id=0):
        """
        Get all IDs in the medline_xml table thare are
        greater than min_id
        :param min_id: the minimum id
        :return: list of ids in ascending order
        """
        sql_query = "SELECT id from medline_xml where id>%d order by id" % min_id
        rows = self.fetchall(sql_query)
        return [row['id'] for row in rows]

    def medline_xml_select_by_pmid(self, pmid):
        """
        fetch row with PMID
        :param pmid:
        :return: row (dict)
        """
        sql_query = "SELECT * FROM medline_xml WHERE pmid=%s" % str(pmid)
        return self.fetchrow(sql_query)

    def medline_xml_select_by_pmid_and_max_tstamp(self, pmid, max_tstamp):
        """
        fetch row with PMID and max Tstamp
        :param pmid (int or str)
        :param max_tstamp (date)
        :return: row (dict)
        """
        sql_query = 'SELECT * FROM medline_xml WHERE pmid=%s and Tstamp>"%s"' % (str(pmid), str(max_tstamp))
        return self.fetchrow(sql_query)


    def medline_xml_select_by_id(self, id):
        """
        fetch row with ID
        :param id: ID of row
        :return: row (dict)
        """
        sql_query = "SELECT * FROM medline_xml WHERE id=%s" % str(id)
        return self.fetchrow(sql_query)

    def medline_xml_update(self, xml, filename_id):
        """
        Insert or update the parsed HGVS variant record

        Will only update if the xml citation has a newer Tstamp
        than the one in the DB

        :param xml: xml (str)
        :return: result of sql query for insertion or update, or None of nothing was inserted
        """
        pmid = self._get_PMID(xml)
        last_date = self._get_last_date(xml)

        # find any row with the same PMID and that are older than max stamp
        record_in_db = self.medline_xml_select_by_pmid_and_max_tstamp(pmid, last_date)

        if record_in_db is None:
            d = {"PMID" : pmid, "xml" : xml, "Tstamp" : str(last_date), "medline_xml_filename_id" : filename_id}
            sql_insert = 'INSERT INTO medline_xml (PMID, xml, Tstamp, medline_xml_filename_id) VALUES (%(PMID)s, %(xml)s, %(Tstamp)s, %(medline_xml_filename_id)s) '
            sql_insert += 'ON DUPLICATE KEY UPDATE xml=values(xml), Tstamp=values(Tstamp), medline_xml_filename_id=values(medline_xml_filename_id)'
            return self.execute(sql_insert, d)

        return None





"""BooleanQueryExpresions for rewriting input to 
multiple different queryable targets such as Google, PubGet, CodedDatabses, etc. 
"""

#TODO: @nthmost: refactor
class BooleanQueryExpression:
    """
    Boolean query expressions can be executed against any target, for example Google or a SQL Database.
    Typically the BooleanQuery is used against a GoogleQuery target.

    Default Target outputs Boolean operators.
    The 'AND' case is implicit
    """
    def __init__(self, _and=" ", _or="|", _equal='=', _quote='"', _open="(", _close=")", _tic="'", _space=' ',):
        self._AND    = _and
        self._OR     = _or
        self._OPEN   = _open
        self._CLOSE  = _close
        self._SPACE  = _space
        self._TIC    = _tic
        self._QUOTE  = _quote
        self._EQUAL  = _equal

    def AND(self, iterable):
        return self.enclose(iterable, self._AND)

    def OR(self, iterable):
        return self.enclose(iterable, self._OR)

    def enclose(self, iterable, AndOr):
        if iterable is None:
            return self._SPACE
        elif type(iterable) is list:
            or_expansion = AndOr.join(self.clean(item) for item in self.unique(iterable))
            return self.parenthesize(or_expansion)
        else:
            return iterable

    def parenthesize(self, iterable):
        return self._OPEN + iterable + self._CLOSE

    #TODO: refactor
    def unique(self, iterable):
        if True:
            return iterable
        else:
            unique = []
            for node in iterable:
                if node not in unique:
                    unique.append(node)
            return unique

    def atleast1(self, expansion):
        if expansion is None:
            return self._SPACE
        elif type(expansion) is list:
            union_of_expansion = (self._OR.join(self.exactly(item) for item in self.unique(expansion)))
            return self.parenthesize(union_of_expansion)
        else:
            return self.exactly(expansion)

    def quote(self, text, quotemark=None):

        if quotemark==None: quotemark=self._QUOTE

        if text is None:
            return quotemark + quotemark

        elif type(text) is list:
            return [(self.quote(item, quotemark)) for item in text]

        else:
            return quotemark +str(text)+ quotemark

    def tic(self, quotable):
        if quotable is None:
            return self._TIC+self._TIC

        elif type(quotable) is list:
            return [(self.tic(item)) for item in quotable]

        else:
            return self._TIC +str(quotable)+ self._TIC

    def equals(self, one, other, exactly=True):
        if exactly:
            return one + self._EQUAL + self.tic(other)
        else:
            return one + self._EQUAL + other

    ############################################################################            
    # Side effects of handling bizarre HGVS rewrite requirements.
    # Fixable. 

    def exactly(self, quotable):
        return self.quote(self.clean(quotable))
        # return '"%s"' % self.clean(quotable)

    #TODO: refactor
    def filter(self, iterable):
        return iterable
        # return filter(None, iterable)

    #TODO: refactor
    def clean(self, expression):
        return self.filter(expression).\
            replace(self._OPEN,'').\
            replace(self._CLOSE,'').\
            replace(self._SPACE+self._SPACE, self._SPACE)



NCBI_GeneID =   "NCBI:GeneID"
HGNC_GeneName = "HGNC:Symbol"
UMLS_ConceptID  = "UMLS:CUI"
MedGen_ConceptID= "MedGen:UID"

class Vocab:
    """
    Annotator Vocabulary source and type of concept.

    examples: ncbi:GeneID, hgnc:GeneName, snomed:ClinicalTerm
    """
    def __init__(self, qualified_label, entry=None):

        self.qualified_label = qualified_label
        self.entry = entry

    def __str__(self):
        if self.entry:
            return str({self.qualified_label:self.entry})
        else:
            return str(self.qualified_label)


import subprocess
from setuptools import setup, find_packages
import medgen

from setuptools.command.build_ext import build_ext as _build_ext

class build_ext(_build_ext):
    # http://stackoverflow.com/questions/21605927/why-doesnt-setup-requires-work-properly-for-numpy/21621493
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setup(
    name = 'medgen-prime',
    version = medgen.__version__,
    description = 'MedGen is the NCBI primary source for Medical Genomics information',
    long_description = 'Modules in medgen.* provide access to the databases and concepts found in medgen-mysql.',
    url = 'https://bitbucket.org/text2gene/medgen',
    author = 'Dr. Andy McMurry',
    maintainer = 'Naomi Most',
    author_email = 'andymc@apache.org',
    maintainer_email = 'naomi@text2gene.com',
    license = 'Apache License 2.0 (http://www.apache.org/licenses/LICENSE-2.0)',
    packages = find_packages(),
    package_data={'medgen': ['config/*.ini']},
    cmdclass = {'build_ext': build_ext},
    setup_requires = ['numpy'],
    install_requires = [
        'setuptools',
        'wheel',
        'configparser',
        'mysql-connector-python',
        'mysqlclient',
        'metapub',
        'numpy',
        'nose',
        'coverage',
        'mock',
        'PyHamcrest',
        'requests',
        'pyrfc3339',
        'ipython',
        ],
    )




##########################################################################################
from ..db.medgen import MedGenDB

##########################################################################################
#
#       Functions
#
##########################################################################################

def _medgen_url(cui):
    """
    URL linkout the NCBI MedGen website.
    :param cui: concept unique identifier
    :return: str url
    """
    return 'http://www.ncbi.nlm.nih.gov/medgen/' + str(cui)

def _define_medgen_concept(cui):
    """
    MedGen defined Concept Definitions.
    Source can be either UMLS or MedGen.
    UMLS definitions come from a variety of sources, such as SNOMED, MESH, etc.
    MedGen concept definitions are often from Office of Rare diseases, HPO terms, etc.

    http://www.ncbi.nlm.nih.gov/books/NBK159970/#_MedGen_Data_Model_

    :param cui: UMLS unique concept id
    :return: Definition string
    """
    try:
        concept_def = MedGenDB().concept_definition(cui)
        concept_def['url'] = _medgen_url(cui)

        return concept_def
    except:
        return None

##########################################################################################
#
#       API
#
##########################################################################################

ConceptName = MedGenDB().concept_name
ConceptDefinition  = _define_medgen_concept
ConceptRelations   = MedGenDB().concept_relations
ConceptSources     = MedGenDB().concept_sources
ConceptURL         = _medgen_url

# ALIAS
Define  = ConceptDefinition
Relate  = ConceptRelations



""" Variant-level annotation functions requiring ClinvarDB and Metapub (NCBI/eutils). """

import requests, json, urllib

from metapub.text_mining import is_pmcid, is_ncbi_bookID 
from metapub.pubmedcentral import get_pmid_for_otherid

from ..db.clinvar import ClinVarDB
from ..log import log

##########################################################################################
#
#   Functions
#
##########################################################################################

def _clinvar_variant_accession(hgvs_text):
    """
    See ClinVar FAQ http://www.ncbi.nlm.nih.gov/clinvar/docs/faq/#accs
    :param hgvs_text: c.DNA
    :return: RCVAccession "Reference ClinVar Accession"
    """
    try:
        return ClinVarDB().accession_for_hgvs_text(str(hgvs_text))
    except Exception as err:
        log.debug("no clinvar accession for variant hgvs_text %s " % hgvs_text)

def _clinvar_variant_allele_id(hgvs_text):
    """
    Get the unique AlleleID
    :param hgvs_text: c.DNA
    :return: AlleleID
    """
    try:
        return ClinVarDB().allele_id_for_hgvs_text(hgvs_text)
    except Exception as err:
        log.debug('no clinvar AlleleID for variant hgvs_text %s ' % hgvs_text)

def _clinvar_variant_variation_id(hgvs_text):
    """
    Get the unique VariationID
    :param hgvs_text: c.DNA
    :return: VariationID
    """
    try:
        return ClinVarDB().variation_id_for_hgvs_text(hgvs_text)
    except Exception as err:
        log.debug('no clinvar VariationID for variant hgvs_text %s ' % hgvs_text)

def _clinvar_variant2pubmed(hgvs_text):
    """
    Get PMID for clinvar variants using the AlleleID key.

    Keep GeneReviews book references (NKBxxxx) without argument.

    ONE EXPENSIVE LOOKUP HERE:
    If the citation_source is PubMedCentral, first convert responses to PMID.

    :param hgvs_text: c.DNA
    :return: set(PMIDs and possibly also NBK ids)
    """
    pubmeds = []
    citations = ClinVarDB().var_citations(hgvs_text)
    if citations:
        for cite in citations:
            some_id = cite['citation_id']

            if is_ncbi_bookID(some_id):
                # Todo: convert? drop??
                pubmeds.append(some_id)
            elif is_pmcid(some_id):
                try:
                    pmid = get_pmid_for_otherid(some_id)
                    if pmid is not None:
                        log.debug('found PubMedCentral PMCID %s, converted to PMID %s ', some_id, str(pmid))
                        pubmeds.append(pmid)
                    else:
                        log.debug('PMID not found for PMCID %s; discarding.', some_id)
                except Exception as err:
                    log.debug('error converting PMCID %s: %r', some_id, err)
            elif cite['citation_source'] == 'PubMed':
                pubmeds.append(some_id)

    #return set([int(entry) for entry in pubmeds])
    return set(pubmeds)


def clinvar2pmid_with_accessions(hgvs_list):
    ret = []
    citations = ClinVarDB().var_citations(hgvs_list)
    if citations:
        for cite in citations:
            article_id = cite['citation_id']
            if is_ncbi_bookID(article_id):
                pmid = article_id
            else:
                pmid = article_id if cite['citation_source'] == 'PubMed' else get_pmid_for_otherid(article_id)
            if pmid:
                ret.append({"hgvs_text": cite['HGVS'], "pmid": pmid, "accession": cite['RCVaccession']})
    return ret


##########################################################################################
#
# API
#
##########################################################################################

ClinvarAccession   = _clinvar_variant_accession
ClinvarAlleleID    = _clinvar_variant_allele_id
ClinvarPubmeds     = _clinvar_variant2pubmed
ClinvarVariationID = _clinvar_variant_variation_id
